from . import lugo
from ..protos import physics_pb2
from math import hypot


def new_vector(from_point: lugo.Point, to_point: lugo.Point):
    v = physics_pb2.Vector()
    v.x = to_point.x - from_point.x
    v.y = to_point.y - from_point.y
    if is_invalid_vector(v):
        raise RuntimeError("A vector cannot have zero length")
    return v


def normalize(v: lugo.Vector):
    length = get_length(v)
    return get_scaled_vector(v, 100 / length)


def get_length(v: lugo.Vector):
    return hypot(v.x, v.y)


def get_scaled_vector(v: lugo.Vector, scale: float):
    if scale <= 0:
        raise RuntimeError("Vector cannot have zero length")
    v2 = physics_pb2.Vector()
    v2.x = v.x * scale
    v2.y = v.y * scale
    return v2


def sub_vector(original_vector: lugo.Vector, sub_vector: lugo.Vector) -> lugo.Vector:
    new_x = original_vector.x - sub_vector.x
    new_y = original_vector.y - sub_vector.y

    new_vector = physics_pb2.Vector()
    new_vector.x = new_x
    new_vector.y = new_y

    if is_invalid_vector(new_vector):
        raise ValueError("Could not subtract vectors: the result would be a zero-length vector")
    return new_vector


def is_invalid_vector(v: lugo.Vector) -> bool:
    return v.x == 0 and v.y == 0


def distance_between_points(a: lugo.Point, b: lugo.Point):
    return hypot(a.x - b.x, a.y - b.y)



from ...protos.physics_pb2 import Point

from ...src.specs import *
from ...src.goal import Goal
from ...protos import server_pb2

class Direction(object):
    pass


DIRECTION = Direction()
DIRECTION.FORWARD = 0
DIRECTION.BACKWARD = 1,
DIRECTION.LEFT = 2,
DIRECTION.RIGHT = 3,
DIRECTION.BACKWARD_LEFT = 4,
DIRECTION.BACKWARD_RIGHT = 5,
DIRECTION.FORWARD_LEFT = 6,
DIRECTION.FORWARD_RIGHT = 7

homeGoalCenter = Point()
homeGoalCenter.x = 0
homeGoalCenter.y = int(MAX_Y_COORDINATE / 2)

homeGoalTopPole = Point()
homeGoalTopPole.x = 0
homeGoalTopPole.y = int(GOAL_MAX_Y)

homeGoalBottomPole = Point()
homeGoalBottomPole.x = 0
homeGoalBottomPole.y = int(GOAL_MIN_Y)

awayGoalCenter = Point()
awayGoalCenter.x = int(MAX_X_COORDINATE)
awayGoalCenter.y = int(MAX_Y_COORDINATE / 2)

awayGoalTopPole = Point()
awayGoalTopPole.x = int(MAX_X_COORDINATE)
awayGoalTopPole.y = int(GOAL_MAX_Y)

awayGoalBottomPole = Point()
awayGoalBottomPole.x = int(MAX_X_COORDINATE)
awayGoalBottomPole.y = int(GOAL_MIN_Y)

awayGoal = Goal(
    server_pb2.Team.Side.AWAY,
    awayGoalCenter,
    awayGoalTopPole,
    awayGoalBottomPole
)
homeGoal = Goal(
    server_pb2.Team.Side.HOME,
    homeGoalCenter,
    homeGoalTopPole,
    homeGoalBottomPole
)


from .src.mapper import Mapper, Region, ERR_MIN_COLS, ERR_MAX_COLS, ERR_MIN_ROWS, ERR_MAX_ROWS, MIN_COLS, MIN_ROWS, \
    MAX_COLS, MAX_ROWS


from .src.orientation import Orientation, ORIENTATION

from .src.direction import Direction, DIRECTION, homeGoalCenter, homeGoalTopPole, homeGoalBottomPole, awayGoalCenter, \
    awayGoalTopPole, awayGoalBottomPole, awayGoal, homeGoal


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: server.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import physics_pb2 as physics__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0cserver.proto\x12\x04lugo\x1a\rphysics.proto\"\x8e\x01\n\x0bJoinRequest\x12\r\n\x05token\x18\x01 \x01(\t\x12\x18\n\x10protocol_version\x18\x02 \x01(\t\x12\"\n\tteam_side\x18\x03 \x01(\x0e\x32\x0f.lugo.Team.Side\x12\x0e\n\x06number\x18\x04 \x01(\r\x12\"\n\rinit_position\x18\x05 \x01(\x0b\x32\x0b.lugo.Point\"\xbc\x02\n\x0cGameSnapshot\x12\'\n\x05state\x18\x01 \x01(\x0e\x32\x18.lugo.GameSnapshot.State\x12\x0c\n\x04turn\x18\x02 \x01(\r\x12\x1d\n\thome_team\x18\x03 \x01(\x0b\x32\n.lugo.Team\x12\x1d\n\taway_team\x18\x04 \x01(\x0b\x32\n.lugo.Team\x12\x18\n\x04\x62\x61ll\x18\x05 \x01(\x0b\x32\n.lugo.Ball\x12\x1f\n\x17turns_ball_in_goal_zone\x18\x06 \x01(\r\x12#\n\nshot_clock\x18\x07 \x01(\x0b\x32\x0f.lugo.ShotClock\"W\n\x05State\x12\x0b\n\x07WAITING\x10\x00\x12\r\n\tGET_READY\x10\x01\x12\r\n\tLISTENING\x10\x02\x12\x0b\n\x07PLAYING\x10\x03\x12\x0c\n\x08SHIFTING\x10\x04\x12\x08\n\x04OVER\x10\x63\"}\n\x04Team\x12\x1d\n\x07players\x18\x01 \x03(\x0b\x32\x0c.lugo.Player\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\r\n\x05score\x18\x03 \x01(\r\x12\x1d\n\x04side\x18\x04 \x01(\x0e\x32\x0f.lugo.Team.Side\"\x1a\n\x04Side\x12\x08\n\x04HOME\x10\x00\x12\x08\n\x04\x41WAY\x10\x01\"H\n\tShotClock\x12\"\n\tteam_side\x18\x06 \x01(\x0e\x32\x0f.lugo.Team.Side\x12\x17\n\x0fremaining_turns\x18\x07 \x01(\r\"\xa1\x01\n\x06Player\x12\x0e\n\x06number\x18\x01 \x01(\r\x12\x1d\n\x08position\x18\x02 \x01(\x0b\x32\x0b.lugo.Point\x12 \n\x08velocity\x18\x03 \x01(\x0b\x32\x0e.lugo.Velocity\x12\"\n\tteam_side\x18\x04 \x01(\x0e\x32\x0f.lugo.Team.Side\x12\"\n\rinit_position\x18\x05 \x01(\x0b\x32\x0b.lugo.Point\"e\n\x04\x42\x61ll\x12\x1d\n\x08position\x18\x01 \x01(\x0b\x32\x0b.lugo.Point\x12 \n\x08velocity\x18\x02 \x01(\x0b\x32\x0e.lugo.Velocity\x12\x1c\n\x06holder\x18\x03 \x01(\x0b\x32\x0c.lugo.Player\"\xab\x01\n\rOrderResponse\x12,\n\x04\x63ode\x18\x01 \x01(\x0e\x32\x1e.lugo.OrderResponse.StatusCode\x12\x0f\n\x07\x64\x65tails\x18\x02 \x01(\t\"[\n\nStatusCode\x12\x0b\n\x07SUCCESS\x10\x00\x12\x12\n\x0eUNKNOWN_PLAYER\x10\x01\x12\x11\n\rNOT_LISTENING\x10\x02\x12\x0e\n\nWRONG_TURN\x10\x03\x12\t\n\x05OTHER\x10\x63\"L\n\x08OrderSet\x12\x0c\n\x04turn\x18\x01 \x01(\r\x12\x1b\n\x06orders\x18\x02 \x03(\x0b\x32\x0b.lugo.Order\x12\x15\n\rdebug_message\x18\x03 \x01(\t\"\x83\x01\n\x05Order\x12\x1a\n\x04move\x18\x01 \x01(\x0b\x32\n.lugo.MoveH\x00\x12\x1c\n\x05\x63\x61tch\x18\x02 \x01(\x0b\x32\x0b.lugo.CatchH\x00\x12\x1a\n\x04kick\x18\x03 \x01(\x0b\x32\n.lugo.KickH\x00\x12\x1a\n\x04jump\x18\x04 \x01(\x0b\x32\n.lugo.JumpH\x00\x42\x08\n\x06\x61\x63tion\"(\n\x04Move\x12 \n\x08velocity\x18\x01 \x01(\x0b\x32\x0e.lugo.Velocity\"\x07\n\x05\x43\x61tch\"(\n\x04Kick\x12 \n\x08velocity\x18\x01 \x01(\x0b\x32\x0e.lugo.Velocity\"(\n\x04Jump\x12 \n\x08velocity\x18\x01 \x01(\x0b\x32\x0e.lugo.Velocity2o\n\x04Game\x12\x34\n\tJoinATeam\x12\x11.lugo.JoinRequest\x1a\x12.lugo.GameSnapshot0\x01\x12\x31\n\nSendOrders\x12\x0e.lugo.OrderSet\x1a\x13.lugo.OrderResponseB#Z!github.com/lugobots/lugo4go/protob\x06proto3')



_JOINREQUEST = DESCRIPTOR.message_types_by_name['JoinRequest']
_GAMESNAPSHOT = DESCRIPTOR.message_types_by_name['GameSnapshot']
_TEAM = DESCRIPTOR.message_types_by_name['Team']
_SHOTCLOCK = DESCRIPTOR.message_types_by_name['ShotClock']
_PLAYER = DESCRIPTOR.message_types_by_name['Player']
_BALL = DESCRIPTOR.message_types_by_name['Ball']
_ORDERRESPONSE = DESCRIPTOR.message_types_by_name['OrderResponse']
_ORDERSET = DESCRIPTOR.message_types_by_name['OrderSet']
_ORDER = DESCRIPTOR.message_types_by_name['Order']
_MOVE = DESCRIPTOR.message_types_by_name['Move']
_CATCH = DESCRIPTOR.message_types_by_name['Catch']
_KICK = DESCRIPTOR.message_types_by_name['Kick']
_JUMP = DESCRIPTOR.message_types_by_name['Jump']
_GAMESNAPSHOT_STATE = _GAMESNAPSHOT.enum_types_by_name['State']
_TEAM_SIDE = _TEAM.enum_types_by_name['Side']
_ORDERRESPONSE_STATUSCODE = _ORDERRESPONSE.enum_types_by_name['StatusCode']
JoinRequest = _reflection.GeneratedProtocolMessageType('JoinRequest', (_message.Message,), {
  'DESCRIPTOR' : _JOINREQUEST,
  '__module__' : 'server_pb2'
  # @@protoc_insertion_point(class_scope:lugo.JoinRequest)
  })
_sym_db.RegisterMessage(JoinRequest)

GameSnapshot = _reflection.GeneratedProtocolMessageType('GameSnapshot', (_message.Message,), {
  'DESCRIPTOR' : _GAMESNAPSHOT,
  '__module__' : 'server_pb2'
  # @@protoc_insertion_point(class_scope:lugo.GameSnapshot)
  })
_sym_db.RegisterMessage(GameSnapshot)

Team = _reflection.GeneratedProtocolMessageType('Team', (_message.Message,), {
  'DESCRIPTOR' : _TEAM,
  '__module__' : 'server_pb2'
  # @@protoc_insertion_point(class_scope:lugo.Team)
  })
_sym_db.RegisterMessage(Team)

ShotClock = _reflection.GeneratedProtocolMessageType('ShotClock', (_message.Message,), {
  'DESCRIPTOR' : _SHOTCLOCK,
  '__module__' : 'server_pb2'
  # @@protoc_insertion_point(class_scope:lugo.ShotClock)
  })
_sym_db.RegisterMessage(ShotClock)

Player = _reflection.GeneratedProtocolMessageType('Player', (_message.Message,), {
  'DESCRIPTOR' : _PLAYER,
  '__module__' : 'server_pb2'
  # @@protoc_insertion_point(class_scope:lugo.Player)
  })
_sym_db.RegisterMessage(Player)

Ball = _reflection.GeneratedProtocolMessageType('Ball', (_message.Message,), {
  'DESCRIPTOR' : _BALL,
  '__module__' : 'server_pb2'
  # @@protoc_insertion_point(class_scope:lugo.Ball)
  })
_sym_db.RegisterMessage(Ball)

OrderResponse = _reflection.GeneratedProtocolMessageType('OrderResponse', (_message.Message,), {
  'DESCRIPTOR' : _ORDERRESPONSE,
  '__module__' : 'server_pb2'
  # @@protoc_insertion_point(class_scope:lugo.OrderResponse)
  })
_sym_db.RegisterMessage(OrderResponse)

OrderSet = _reflection.GeneratedProtocolMessageType('OrderSet', (_message.Message,), {
  'DESCRIPTOR' : _ORDERSET,
  '__module__' : 'server_pb2'
  # @@protoc_insertion_point(class_scope:lugo.OrderSet)
  })
_sym_db.RegisterMessage(OrderSet)

Order = _reflection.GeneratedProtocolMessageType('Order', (_message.Message,), {
  'DESCRIPTOR' : _ORDER,
  '__module__' : 'server_pb2'
  # @@protoc_insertion_point(class_scope:lugo.Order)
  })
_sym_db.RegisterMessage(Order)

Move = _reflection.GeneratedProtocolMessageType('Move', (_message.Message,), {
  'DESCRIPTOR' : _MOVE,
  '__module__' : 'server_pb2'
  # @@protoc_insertion_point(class_scope:lugo.Move)
  })
_sym_db.RegisterMessage(Move)

Catch = _reflection.GeneratedProtocolMessageType('Catch', (_message.Message,), {
  'DESCRIPTOR' : _CATCH,
  '__module__' : 'server_pb2'
  # @@protoc_insertion_point(class_scope:lugo.Catch)
  })
_sym_db.RegisterMessage(Catch)

Kick = _reflection.GeneratedProtocolMessageType('Kick', (_message.Message,), {
  'DESCRIPTOR' : _KICK,
  '__module__' : 'server_pb2'
  # @@protoc_insertion_point(class_scope:lugo.Kick)
  })
_sym_db.RegisterMessage(Kick)

Jump = _reflection.GeneratedProtocolMessageType('Jump', (_message.Message,), {
  'DESCRIPTOR' : _JUMP,
  '__module__' : 'server_pb2'
  # @@protoc_insertion_point(class_scope:lugo.Jump)
  })
_sym_db.RegisterMessage(Jump)

_GAME = DESCRIPTOR.services_by_name['Game']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z!github.com/lugobots/lugo4go/proto'
  _JOINREQUEST._serialized_start=38
  _JOINREQUEST._serialized_end=180
  _GAMESNAPSHOT._serialized_start=183
  _GAMESNAPSHOT._serialized_end=499
  _GAMESNAPSHOT_STATE._serialized_start=412
  _GAMESNAPSHOT_STATE._serialized_end=499
  _TEAM._serialized_start=501
  _TEAM._serialized_end=626
  _TEAM_SIDE._serialized_start=600
  _TEAM_SIDE._serialized_end=626
  _SHOTCLOCK._serialized_start=628
  _SHOTCLOCK._serialized_end=700
  _PLAYER._serialized_start=703
  _PLAYER._serialized_end=864
  _BALL._serialized_start=866
  _BALL._serialized_end=967
  _ORDERRESPONSE._serialized_start=970
  _ORDERRESPONSE._serialized_end=1141
  _ORDERRESPONSE_STATUSCODE._serialized_start=1050
  _ORDERRESPONSE_STATUSCODE._serialized_end=1141
  _ORDERSET._serialized_start=1143
  _ORDERSET._serialized_end=1219
  _ORDER._serialized_start=1222
  _ORDER._serialized_end=1353
  _MOVE._serialized_start=1355
  _MOVE._serialized_end=1395
  _CATCH._serialized_start=1397
  _CATCH._serialized_end=1404
  _KICK._serialized_start=1406
  _KICK._serialized_end=1446
  _JUMP._serialized_start=1448
  _JUMP._serialized_end=1488
  _GAME._serialized_start=1490
  _GAME._serialized_end=1601
# @@protoc_insertion_point(module_scope)



from . import lugo
from ..protos.physics_pb2 import Point
from ..protos.server_pb2 import Team


class Goal(object):

    def __init__(self, place: lugo.TeamSide, center: lugo.Point, top_pole: lugo.Point, bottom_pole: lugo.Point):
        self._center = center
        self._place = place
        self._topPole = top_pole
        self._bottomPole = bottom_pole

    def get_center(self) -> Point:
        return self._center

    def get_place(self) -> Team.Side:
        return self._place

    def get_top_pole(self) -> Point:
        return self._topPole

    def get_bottom_pole(self) -> Point:
        return self._bottomPole



from . import lugo, geo, interface, specs
from ..mapper import DIRECTION, ORIENTATION, homeGoal, awayGoal
from .goal import Goal
from ..mapper.src.orientation import NORTH
from ..protos import server_pb2
from ..protos.physics_pb2 import Point


class GameSnapshotReader:
    def __init__(self, snapshot: lugo.GameSnapshot, my_side: lugo.TeamSide):
        self.snapshot = snapshot
        self.my_side = my_side

    def get_my_team(self) -> lugo.Team:
        return self.get_team(self.my_side)

    def get_opponent_team(self) -> lugo.Team:
        return self.get_team(self.get_opponent_side())

    def get_team(self, side) -> lugo.Team:
        if side == server_pb2.Team.Side.HOME:
            return self.snapshot.home_team

        return self.snapshot.away_team

    def is_ball_holder(self, player: lugo.Player) -> bool:
        ball = self.snapshot.ball

        return ball.holder is not None and ball.holder.team_side == player.team_side and ball.holder.number == player.number

    def get_opponent_side(self) -> lugo.TeamSide:
        if self.my_side == server_pb2.Team.Side.HOME:
            return server_pb2.Team.Side.AWAY

        return server_pb2.Team.Side.HOME

    def get_my_goal(self) -> Goal:
        if self.my_side == server_pb2.Team.Side.HOME:
            return homeGoal

        return awayGoal

    def get_ball(self) -> lugo.Ball:
        return self.snapshot.ball

    def get_opponent_goal(self) -> Goal:
        if self.my_side == server_pb2.Team.Side.HOME:
            return awayGoal

        return homeGoal

    def get_player(self, side: server_pb2.Team.Side, number: int) -> lugo.Player:
        team = self.get_team(side)
        if team is None:
            return None

        for player in team.players:
            if player.number == number:
                return player
        return None

    def make_order_move_max_speed(self, origin: lugo.Point, target: lugo.Point) -> lugo.Order:
        return self.make_order_move(origin, target, specs.PLAYER_MAX_SPEED)

    def make_order_move(self, origin: lugo.Point, target: lugo.Point, speed: int) -> lugo.Order:
        if origin.x == target.x and origin.y == target.y:
            # a vector cannot have zeroed direction. In this case, the player will just be stopped
            return self.make_order_move_from_vector(NORTH, 0)

        direction = geo.new_vector(origin, target)
        direction = geo.normalize(direction)
        return self.make_order_move_from_vector(direction, speed)

    def make_order_move_from_vector(self, direction: lugo.Vector, speed: int) -> lugo.Order:
        order = server_pb2.Order()

        order.move.velocity.direction.CopyFrom(direction)
        order.move.velocity.speed = speed
        return order

    def make_order_move_by_direction(self, direction) -> lugo.Order:
        if direction == DIRECTION.FORWARD:
            direction_target = ORIENTATION.EAST
            if self.my_side == server_pb2.Team.Side.AWAY:
                direction_target = ORIENTATION.WEST

        elif direction == DIRECTION.BACKWARD:
            direction_target = ORIENTATION.WEST
            if self.my_side == server_pb2.Team.Side.AWAY:
                direction_target = ORIENTATION.EAST

        elif direction == DIRECTION.LEFT:
            direction_target = ORIENTATION.NORTH
            if self.my_side == server_pb2.Team.Side.AWAY:
                direction_target = ORIENTATION.SOUTH

        elif direction == DIRECTION.RIGHT:
            direction_target = ORIENTATION.SOUTH
            if self.my_side == server_pb2.Team.Side.AWAY:
                direction_target = ORIENTATION.NORTH

        elif direction == DIRECTION.BACKWARD_LEFT:
            direction_target = ORIENTATION.NORTH_WEST
            if self.my_side == server_pb2.Team.Side.AWAY:
                direction_target = ORIENTATION.SOUTH_EAST

        elif direction == DIRECTION.BACKWARD_RIGHT:
            direction_target = ORIENTATION.SOUTH_WEST
            if self.my_side == server_pb2.Team.Side.AWAY:
                direction_target = ORIENTATION.NORTH_EAST

        elif direction == DIRECTION.FORWARD_LEFT:
            direction_target = ORIENTATION.NORTH_EAST
            if self.my_side == server_pb2.Team.Side.AWAY:
                direction_target = ORIENTATION.SOUTH_WEST

        elif direction == DIRECTION.FORWARD_RIGHT:
            direction_target = ORIENTATION.SOUTH_EAST
            if self.my_side == server_pb2.Team.Side.AWAY:
                direction_target = ORIENTATION.NORTH_WEST

        else:
            raise AttributeError('unknown direction {direction}')

        return self.make_order_move_from_vector(direction_target, specs.PLAYER_MAX_SPEED)

    def make_order_jump(self, origin: lugo.Point, target: lugo.Point, speed: int) -> lugo.Order:
        direction = ORIENTATION.EAST
        if origin.x != target.x or origin.y != target.y:
            # a vector cannot have zeroed direction. In this case, the player will just be stopped
            direction = geo.new_vector(origin, target)
            direction = geo.normalize(direction)

        new_velocity = lugo.new_velocity(direction)
        new_velocity.speed = speed

        order = server_pb2.Order()
        jump = order.jump
        jump.velocity.CopyFrom(new_velocity)

        return order

    def make_order_kick(self, ball: lugo.Ball, target: Point, speed: int) -> lugo.Order:
        ball_expected_direction = geo.new_vector(ball.position, target)

        # the ball velocity is summed to the kick velocity, so we have to consider the current ball direction
        diff_vector = geo.sub_vector(
            ball_expected_direction, ball.velocity.direction)

        new_velocity = lugo.new_velocity(geo.normalize(diff_vector))
        new_velocity.speed = speed

        order = server_pb2.Order()
        order.kick.velocity.CopyFrom(new_velocity)

        return order

    def make_order_kick_max_speed(self, ball: lugo.Ball, target: Point) -> lugo.Order:
        return self.make_order_kick(ball, target, specs.BALL_MAX_SPEED)

    def make_order_catch(self) -> server_pb2.Order:
        order = server_pb2.Order()
        order.catch.SetInParent()
        return order





def define_state(snapshot: lugo.GameSnapshot, player_number: int, side: lugo.TeamSide) -> interface.PLAYER_STATE:
    if not snapshot or not snapshot.ball:
        raise AttributeError(
            'invalid snapshot state - cannot define player state')

    reader = GameSnapshotReader(snapshot, side)
    me = reader.get_player(side, player_number)
    if me is None:
        raise AttributeError(
            'could not find the bot in the snapshot - cannot define player state')

    ball_holder = snapshot.ball.holder

    if ball_holder.number == 0:
        return interface.PLAYER_STATE.DISPUTING_THE_BALL

    if ball_holder.team_side == side:
        if ball_holder.number == player_number:
            return interface.PLAYER_STATE.HOLDING_THE_BALL

        return interface.PLAYER_STATE.SUPPORTING

    return interface.PLAYER_STATE.DEFENDING



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import remote_pb2 as remote__pb2


class RemoteStub(object):
    """The game server implements a Remote service that allows you to control the game flow.
    This service may help you to control or watch the game during training sessions.
    The game server only offers this service on debug mode on.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.PauseOrResume = channel.unary_unary(
                '/lugo.Remote/PauseOrResume',
                request_serializer=remote__pb2.PauseResumeRequest.SerializeToString,
                response_deserializer=remote__pb2.CommandResponse.FromString,
                )
        self.NextTurn = channel.unary_unary(
                '/lugo.Remote/NextTurn',
                request_serializer=remote__pb2.NextTurnRequest.SerializeToString,
                response_deserializer=remote__pb2.CommandResponse.FromString,
                )
        self.NextOrder = channel.unary_unary(
                '/lugo.Remote/NextOrder',
                request_serializer=remote__pb2.NextOrderRequest.SerializeToString,
                response_deserializer=remote__pb2.CommandResponse.FromString,
                )
        self.SetBallProperties = channel.unary_unary(
                '/lugo.Remote/SetBallProperties',
                request_serializer=remote__pb2.BallProperties.SerializeToString,
                response_deserializer=remote__pb2.CommandResponse.FromString,
                )
        self.SetPlayerProperties = channel.unary_unary(
                '/lugo.Remote/SetPlayerProperties',
                request_serializer=remote__pb2.PlayerProperties.SerializeToString,
                response_deserializer=remote__pb2.CommandResponse.FromString,
                )
        self.SetGameProperties = channel.unary_unary(
                '/lugo.Remote/SetGameProperties',
                request_serializer=remote__pb2.GameProperties.SerializeToString,
                response_deserializer=remote__pb2.CommandResponse.FromString,
                )
        self.ResumeListeningPhase = channel.unary_unary(
                '/lugo.Remote/ResumeListeningPhase',
                request_serializer=remote__pb2.ResumeListeningRequest.SerializeToString,
                response_deserializer=remote__pb2.ResumeListeningResponse.FromString,
                )


class RemoteServicer(object):
    """The game server implements a Remote service that allows you to control the game flow.
    This service may help you to control or watch the game during training sessions.
    The game server only offers this service on debug mode on.
    """

    def PauseOrResume(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def NextTurn(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def NextOrder(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetBallProperties(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetPlayerProperties(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetGameProperties(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ResumeListeningPhase(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_RemoteServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'PauseOrResume': grpc.unary_unary_rpc_method_handler(
                    servicer.PauseOrResume,
                    request_deserializer=remote__pb2.PauseResumeRequest.FromString,
                    response_serializer=remote__pb2.CommandResponse.SerializeToString,
            ),
            'NextTurn': grpc.unary_unary_rpc_method_handler(
                    servicer.NextTurn,
                    request_deserializer=remote__pb2.NextTurnRequest.FromString,
                    response_serializer=remote__pb2.CommandResponse.SerializeToString,
            ),
            'NextOrder': grpc.unary_unary_rpc_method_handler(
                    servicer.NextOrder,
                    request_deserializer=remote__pb2.NextOrderRequest.FromString,
                    response_serializer=remote__pb2.CommandResponse.SerializeToString,
            ),
            'SetBallProperties': grpc.unary_unary_rpc_method_handler(
                    servicer.SetBallProperties,
                    request_deserializer=remote__pb2.BallProperties.FromString,
                    response_serializer=remote__pb2.CommandResponse.SerializeToString,
            ),
            'SetPlayerProperties': grpc.unary_unary_rpc_method_handler(
                    servicer.SetPlayerProperties,
                    request_deserializer=remote__pb2.PlayerProperties.FromString,
                    response_serializer=remote__pb2.CommandResponse.SerializeToString,
            ),
            'SetGameProperties': grpc.unary_unary_rpc_method_handler(
                    servicer.SetGameProperties,
                    request_deserializer=remote__pb2.GameProperties.FromString,
                    response_serializer=remote__pb2.CommandResponse.SerializeToString,
            ),
            'ResumeListeningPhase': grpc.unary_unary_rpc_method_handler(
                    servicer.ResumeListeningPhase,
                    request_deserializer=remote__pb2.ResumeListeningRequest.FromString,
                    response_serializer=remote__pb2.ResumeListeningResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'lugo.Remote', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class Remote(object):
    """The game server implements a Remote service that allows you to control the game flow.
    This service may help you to control or watch the game during training sessions.
    The game server only offers this service on debug mode on.
    """

    @staticmethod
    def PauseOrResume(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lugo.Remote/PauseOrResume',
            remote__pb2.PauseResumeRequest.SerializeToString,
            remote__pb2.CommandResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def NextTurn(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lugo.Remote/NextTurn',
            remote__pb2.NextTurnRequest.SerializeToString,
            remote__pb2.CommandResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def NextOrder(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lugo.Remote/NextOrder',
            remote__pb2.NextOrderRequest.SerializeToString,
            remote__pb2.CommandResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SetBallProperties(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lugo.Remote/SetBallProperties',
            remote__pb2.BallProperties.SerializeToString,
            remote__pb2.CommandResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SetPlayerProperties(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lugo.Remote/SetPlayerProperties',
            remote__pb2.PlayerProperties.SerializeToString,
            remote__pb2.CommandResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SetGameProperties(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lugo.Remote/SetGameProperties',
            remote__pb2.GameProperties.SerializeToString,
            remote__pb2.CommandResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ResumeListeningPhase(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lugo.Remote/ResumeListeningPhase',
            remote__pb2.ResumeListeningRequest.SerializeToString,
            remote__pb2.ResumeListeningResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



from .src.gym import Gym

from .src.helper_bots import newZombieHelperPlayer, newChaserHelperPlayer, newCustomHelperPlayer

from .src.interfaces import TrainingController, BotTrainer


from .src.remote_control import RemoteControl

from .src.training_controller import TrainingCrl


from ...protos import physics_pb2
from ...src import geo

east_vector = physics_pb2.Vector()
east_vector.x = 1
east_vector.y = 0
west_vector = physics_pb2.Vector()
west_vector.x = -1
west_vector.y = 0
south_vector = physics_pb2.Vector()
south_vector.x = 0
south_vector.y = -1
north_vector = physics_pb2.Vector()
north_vector.x = 0
north_vector.y = 1
northeast_vector = physics_pb2.Vector()
northeast_vector.x = 1
northeast_vector.y = 1
northwest_vector = physics_pb2.Vector()
northwest_vector.x = -1
northwest_vector.y = 1
southeast_vector = physics_pb2.Vector()
southeast_vector.x = 1
southeast_vector.y = -1
southwest_vector = physics_pb2.Vector()
southwest_vector.x = -1
southwest_vector.y = -1

EAST = geo.normalize(east_vector)
WEST = geo.normalize(west_vector)
SOUTH = geo.normalize(south_vector)
NORTH = geo.normalize(north_vector)

NORTH_EAST = geo.normalize(northeast_vector)
NORTH_WEST = geo.normalize(northwest_vector)
SOUTH_EAST = geo.normalize(southeast_vector)
SOUTH_WEST = geo.normalize(southwest_vector)

class Orientation(object):
    pass

ORIENTATION = Orientation()
ORIENTATION.EAST = EAST
ORIENTATION.WEST = WEST
ORIENTATION.SOUTH = SOUTH
ORIENTATION.NORTH = NORTH
ORIENTATION.NORTH_EAST = NORTH_EAST
ORIENTATION.NORTH_WEST = NORTH_WEST
ORIENTATION.SOUTH_EAST = SOUTH_EAST
ORIENTATION.SOUTH_WEST = SOUTH_WEST


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: remote.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import physics_pb2 as physics__pb2
from . import server_pb2 as server__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0cremote.proto\x12\x04lugo\x1a\rphysics.proto\x1a\x0cserver.proto\"\x14\n\x12PauseResumeRequest\"\x11\n\x0fNextTurnRequest\"\x12\n\x10NextOrderRequest\"o\n\x0e\x42\x61llProperties\x12\x1d\n\x08position\x18\x01 \x01(\x0b\x32\x0b.lugo.Point\x12 \n\x08velocity\x18\x02 \x01(\x0b\x32\x0e.lugo.Velocity\x12\x1c\n\x06holder\x18\x03 \x01(\x0b\x32\x0c.lugo.Player\"\x82\x01\n\x10PlayerProperties\x12\x1d\n\x04side\x18\x01 \x01(\x0e\x32\x0f.lugo.Team.Side\x12\x0e\n\x06number\x18\x02 \x01(\r\x12\x1d\n\x08position\x18\x03 \x01(\x0b\x32\x0b.lugo.Point\x12 \n\x08velocity\x18\x04 \x01(\x0b\x32\x0e.lugo.Velocity\"\x83\x01\n\x0eGameProperties\x12\x0c\n\x04turn\x18\x01 \x01(\r\x12\x12\n\nhome_score\x18\x02 \x01(\r\x12\x12\n\naway_score\x18\x03 \x01(\r\x12\x16\n\x0e\x66rame_interval\x18\x04 \x01(\x03\x12#\n\nshot_clock\x18\x05 \x01(\x0b\x32\x0f.lugo.ShotClock\"\xcd\x01\n\x0f\x43ommandResponse\x12.\n\x04\x63ode\x18\x01 \x01(\x0e\x32 .lugo.CommandResponse.StatusCode\x12)\n\rgame_snapshot\x18\x02 \x01(\x0b\x32\x12.lugo.GameSnapshot\x12\x0f\n\x07\x64\x65tails\x18\x03 \x01(\t\"N\n\nStatusCode\x12\x0b\n\x07SUCCESS\x10\x00\x12\x11\n\rINVALID_VALUE\x10\x01\x12\x15\n\x11\x44\x45\x41\x44LINE_EXCEEDED\x10\x02\x12\t\n\x05OTHER\x10\x63\"\x18\n\x16ResumeListeningRequest\"\x19\n\x17ResumeListeningResponse2\xdf\x03\n\x06Remote\x12@\n\rPauseOrResume\x12\x18.lugo.PauseResumeRequest\x1a\x15.lugo.CommandResponse\x12\x38\n\x08NextTurn\x12\x15.lugo.NextTurnRequest\x1a\x15.lugo.CommandResponse\x12:\n\tNextOrder\x12\x16.lugo.NextOrderRequest\x1a\x15.lugo.CommandResponse\x12@\n\x11SetBallProperties\x12\x14.lugo.BallProperties\x1a\x15.lugo.CommandResponse\x12\x44\n\x13SetPlayerProperties\x12\x16.lugo.PlayerProperties\x1a\x15.lugo.CommandResponse\x12@\n\x11SetGameProperties\x12\x14.lugo.GameProperties\x1a\x15.lugo.CommandResponse\x12S\n\x14ResumeListeningPhase\x12\x1c.lugo.ResumeListeningRequest\x1a\x1d.lugo.ResumeListeningResponseB#Z!github.com/lugobots/lugo4go/protob\x06proto3')



_PAUSERESUMEREQUEST = DESCRIPTOR.message_types_by_name['PauseResumeRequest']
_NEXTTURNREQUEST = DESCRIPTOR.message_types_by_name['NextTurnRequest']
_NEXTORDERREQUEST = DESCRIPTOR.message_types_by_name['NextOrderRequest']
_BALLPROPERTIES = DESCRIPTOR.message_types_by_name['BallProperties']
_PLAYERPROPERTIES = DESCRIPTOR.message_types_by_name['PlayerProperties']
_GAMEPROPERTIES = DESCRIPTOR.message_types_by_name['GameProperties']
_COMMANDRESPONSE = DESCRIPTOR.message_types_by_name['CommandResponse']
_RESUMELISTENINGREQUEST = DESCRIPTOR.message_types_by_name['ResumeListeningRequest']
_RESUMELISTENINGRESPONSE = DESCRIPTOR.message_types_by_name['ResumeListeningResponse']
_COMMANDRESPONSE_STATUSCODE = _COMMANDRESPONSE.enum_types_by_name['StatusCode']
PauseResumeRequest = _reflection.GeneratedProtocolMessageType('PauseResumeRequest', (_message.Message,), {
  'DESCRIPTOR' : _PAUSERESUMEREQUEST,
  '__module__' : 'remote_pb2'
  # @@protoc_insertion_point(class_scope:lugo.PauseResumeRequest)
  })
_sym_db.RegisterMessage(PauseResumeRequest)

NextTurnRequest = _reflection.GeneratedProtocolMessageType('NextTurnRequest', (_message.Message,), {
  'DESCRIPTOR' : _NEXTTURNREQUEST,
  '__module__' : 'remote_pb2'
  # @@protoc_insertion_point(class_scope:lugo.NextTurnRequest)
  })
_sym_db.RegisterMessage(NextTurnRequest)

NextOrderRequest = _reflection.GeneratedProtocolMessageType('NextOrderRequest', (_message.Message,), {
  'DESCRIPTOR' : _NEXTORDERREQUEST,
  '__module__' : 'remote_pb2'
  # @@protoc_insertion_point(class_scope:lugo.NextOrderRequest)
  })
_sym_db.RegisterMessage(NextOrderRequest)

BallProperties = _reflection.GeneratedProtocolMessageType('BallProperties', (_message.Message,), {
  'DESCRIPTOR' : _BALLPROPERTIES,
  '__module__' : 'remote_pb2'
  # @@protoc_insertion_point(class_scope:lugo.BallProperties)
  })
_sym_db.RegisterMessage(BallProperties)

PlayerProperties = _reflection.GeneratedProtocolMessageType('PlayerProperties', (_message.Message,), {
  'DESCRIPTOR' : _PLAYERPROPERTIES,
  '__module__' : 'remote_pb2'
  # @@protoc_insertion_point(class_scope:lugo.PlayerProperties)
  })
_sym_db.RegisterMessage(PlayerProperties)

GameProperties = _reflection.GeneratedProtocolMessageType('GameProperties', (_message.Message,), {
  'DESCRIPTOR' : _GAMEPROPERTIES,
  '__module__' : 'remote_pb2'
  # @@protoc_insertion_point(class_scope:lugo.GameProperties)
  })
_sym_db.RegisterMessage(GameProperties)

CommandResponse = _reflection.GeneratedProtocolMessageType('CommandResponse', (_message.Message,), {
  'DESCRIPTOR' : _COMMANDRESPONSE,
  '__module__' : 'remote_pb2'
  # @@protoc_insertion_point(class_scope:lugo.CommandResponse)
  })
_sym_db.RegisterMessage(CommandResponse)

ResumeListeningRequest = _reflection.GeneratedProtocolMessageType('ResumeListeningRequest', (_message.Message,), {
  'DESCRIPTOR' : _RESUMELISTENINGREQUEST,
  '__module__' : 'remote_pb2'
  # @@protoc_insertion_point(class_scope:lugo.ResumeListeningRequest)
  })
_sym_db.RegisterMessage(ResumeListeningRequest)

ResumeListeningResponse = _reflection.GeneratedProtocolMessageType('ResumeListeningResponse', (_message.Message,), {
  'DESCRIPTOR' : _RESUMELISTENINGRESPONSE,
  '__module__' : 'remote_pb2'
  # @@protoc_insertion_point(class_scope:lugo.ResumeListeningResponse)
  })
_sym_db.RegisterMessage(ResumeListeningResponse)

_REMOTE = DESCRIPTOR.services_by_name['Remote']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z!github.com/lugobots/lugo4go/proto'
  _PAUSERESUMEREQUEST._serialized_start=51
  _PAUSERESUMEREQUEST._serialized_end=71
  _NEXTTURNREQUEST._serialized_start=73
  _NEXTTURNREQUEST._serialized_end=90
  _NEXTORDERREQUEST._serialized_start=92
  _NEXTORDERREQUEST._serialized_end=110
  _BALLPROPERTIES._serialized_start=112
  _BALLPROPERTIES._serialized_end=223
  _PLAYERPROPERTIES._serialized_start=226
  _PLAYERPROPERTIES._serialized_end=356
  _GAMEPROPERTIES._serialized_start=359
  _GAMEPROPERTIES._serialized_end=490
  _COMMANDRESPONSE._serialized_start=493
  _COMMANDRESPONSE._serialized_end=698
  _COMMANDRESPONSE_STATUSCODE._serialized_start=620
  _COMMANDRESPONSE_STATUSCODE._serialized_end=698
  _RESUMELISTENINGREQUEST._serialized_start=700
  _RESUMELISTENINGREQUEST._serialized_end=724
  _RESUMELISTENINGRESPONSE._serialized_start=726
  _RESUMELISTENINGRESPONSE._serialized_end=751
  _REMOTE._serialized_start=754
  _REMOTE._serialized_end=1233
# @@protoc_insertion_point(module_scope)



import time

from .training_controller import TrainingCrl
from .helper_bots import newChaserHelperPlayer, newZombieHelperPlayer
from .remote_control import RemoteControl
from .interfaces import BotTrainer, TrainingFunction
from ...src import client
from ...protos.server_pb2 import Team
import threading

from concurrent.futures import ThreadPoolExecutor


class Gym:

    def __init__(
            self,
            executor: ThreadPoolExecutor,
            remote_control: RemoteControl,
            trainer: BotTrainer,
            trainingFunction: TrainingFunction,
            options=None,
    ):
        if options is None:
            options = {"debugging_log": False}

        self.remoteControl = remote_control
        self.debugging_log = options["debugging_log"]
        self.trainingCrl = TrainingCrl(executor,
                                       remote_control, trainer, trainingFunction)

        self.trainingCrl.logger = self._debug
        self.gameServerAddress = None
        self.helperPlayers = None
        self.players = []

    def start(self, lugo_client: client.LugoClient, executor: ThreadPoolExecutor):
        hasStarted = False

        def play_callback(orderSet, snapshot):
            nonlocal hasStarted
            hasStarted = True
            return self.trainingCrl.gameTurnHandler(orderSet, snapshot)

        def trigger_listening() -> None:
            nonlocal hasStarted
            if hasStarted is False:
                waiter = threading.Event()
                executor.submit(self.remoteControl.resume_listening, waiter)
                waiter.wait()

        def on_join() -> None:
            self._debug('The main bot is connected!! Starting to connect the zombies')
            time.sleep(0.2)
            if self.gameServerAddress:
                self.players = self.helperPlayers(self.gameServerAddress, executor)
            self._debug('helpers are done')
            trigger_listening()

        lugo_client.play(executor, play_callback, on_join)
        return lugo_client

    def stop(self):
        self.trainingCrl.stop()
        for player in self.players:
            player.stop()

    def with_zombie_players(self, game_server_address):
        self._debug('Entering with_zombie_players\n')
        self.gameServerAddress = game_server_address
        self.helperPlayers = create_helper_players
        return self

    def withChasersPlayers(self, game_server_address):
        self.gameServerAddress = game_server_address

        def helper_players(game_server_address):
            for i in range(1, 12):
                newChaserHelperPlayer(Team.Side.HOME, i, game_server_address)
                newChaserHelperPlayer(Team.Side.AWAY, i, game_server_address)

        self.helperPlayers = helper_players(game_server_address)
        return self

    def _debug(self, message: str):
        if self.debugging_log:
            t = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
            print(f"[Debugger {t}] {message}")


def create_helper_players(gameServerAddress: str, executor: ThreadPoolExecutor):
    children = []
    for i in range(0, 11):
        time.sleep(0.01)
        children.append(newZombieHelperPlayer(Team.Side.HOME, i + 1, gameServerAddress, executor))
        time.sleep(0.01)
        children.append(newZombieHelperPlayer(Team.Side.AWAY, i + 1, gameServerAddress, executor))
    return children



"""
File: lugo.py
Author: Angelo Katipunan
Date: May 20, 2023
Description: This file mocks the gRPC methods to help IDEs intellisense.

Python gRPC files are not friendly to IDEs, what makes the intellisense experience very poor or, sometimes, impossible.

In order to help the programmer experience while developing bots, this file mocks the gRPC methods in a more friendly way.

In short, this file content is not used at all by the package, but will guide the IDE to help the devs.
If you are looking for the real implementation of these methods, please look at the `protos` directory (good luck on that)

"""

from enum import IntEnum
from typing import List

from ..protos import physics_pb2


class Vector:
    def __init__(self, x=0.0, y=0.0):
        self.x = x
        self.y = y


class Point:
    def __init__(self, x=0, y=0):
        self.x = x
        self.y = y


class Velocity:
    def __init__(self, direction=None, speed=0.0):
        self.direction = direction if direction is not None else Vector()
        self.speed = speed


def new_velocity(point: Point) -> Velocity:
    v = physics_pb2.Velocity()
    v.direction.x = point.x
    v.direction.y = point.y
    return v


# Classes from remote.proto

class RemoteServicer(object):

    def PauseOrResume(self, request, context):
        # TODO: Implement the PauseOrResume method.
        response = CommandResponse()
        return response

    def NextTurn(self, request, context):
        # TODO: Implement the NextTurn method.
        response = CommandResponse()
        return response

    def NextOrder(self, request, context):
        # TODO: Implement the NextOrder method.
        response = CommandResponse()
        return response

    def SetBallProperties(self, request, context):
        # TODO: Implement the SetBallProperties method.
        response = CommandResponse()
        return response

    def SetPlayerProperties(self, request, context):
        # TODO: Implement the SetPlayerProperties method.
        response = CommandResponse()
        return response

    def SetGameProperties(self, request, context):
        # TODO: Implement the SetGameProperties method.
        response = CommandResponse()
        return response

    def ResumeListeningPhase(self, request, context):
        # TODO: Implement the ResumeListeningPhase method.
        response = ResumeListeningResponse()
        return response


class PauseResumeRequest(object):
    def __init__(self):
        pass


class NextTurnRequest(object):
    def __init__(self):
        pass


class NextOrderRequest(object):
    def __init__(self):
        pass


class BallProperties:
    def __init__(self, position=None, velocity=None, holder=None):
        self.position = position if position is not None else Point()
        self.velocity = velocity if velocity is not None else Velocity()
        self.holder = holder


class PlayerProperties:
    def __init__(self, side=None, number=0, position=None, velocity=None):
        self.side = side
        self.number = number
        self.position = position if position is not None else Point()
        self.velocity = velocity if velocity is not None else Velocity()


class GameProperties:
    def __init__(self, turn=0, home_score=0, away_score=0, frame_interval=0, shot_clock=None):
        self.turn = turn
        self.home_score = home_score
        self.away_score = away_score
        self.frame_interval = frame_interval
        self.shot_clock = shot_clock


class CommandResponse:
    class StatusCode(IntEnum):
        SUCCESS = 0
        INVALID_VALUE = 1
        DEADLINE_EXCEEDED = 2
        OTHER = 99

    def __init__(self, code=None, game_snapshot=None, details=''):
        self.code = code if code is not None else CommandResponse.StatusCode.SUCCESS
        self.game_snapshot = game_snapshot
        self.details = details


class ResumeListeningRequest(object):
    def __init__(self):
        pass


class ResumeListeningResponse(object):
    def __init__(self):
        pass


# from file src/server.proto


class TeamSide(IntEnum):
    HOME = 0
    AWAY = 1


class State(IntEnum):
    WAITING = 0
    GET_READY = 1
    LISTENING = 2
    PLAYING = 3
    SHIFTING = 4
    OVER = 99


class StatusCode(IntEnum):
    SUCCESS = 0
    UNKNOWN_PLAYER = 1
    NOT_LISTENING = 2
    WRONG_TURN = 3
    OTHER = 99


class Player:
    def __init__(self, number: int, position: Point, velocity: Velocity, team_side: TeamSide, init_position: Point):
        self.number = number

        if position.x is None:
            position.x = 0
        if position.y is None:
            position.y = 0

        self.position = position
        self.velocity = velocity
        self.team_side = team_side
        self.init_position = init_position


class Team:
    def __init__(self, players: List[Player], name: str, score: int, side: TeamSide):
        self.players = players
        self.name = name
        self.score = score
        self.side = side


class Ball:
    def __init__(self, position: Point, velocity: Velocity, holder: Player):
        self.position = position
        self.velocity = velocity
        self.holder = holder


class ShotClock:
    def __init__(self, team_side: TeamSide, remaining_turns: int):
        self.team_side = team_side
        self.remaining_turns = remaining_turns


class JoinRequest:
    def __init__(self, token: str, protocol_version: str, team_side: TeamSide, number: int, init_position: Point):
        self.token = token
        self.protocol_version = protocol_version
        self.team_side = team_side
        self.number = number
        self.init_position = init_position


class GameSnapshot:
    def __init__(self, state: State, turn: int, home_team: Team, away_team: Team, ball: Ball,
                 turns_ball_in_goal_zone: int, shot_clock: ShotClock):
        self.state = state
        self.turn = turn
        self.home_team = home_team
        self.away_team = away_team
        self.ball = ball
        self.turns_ball_in_goal_zone = turns_ball_in_goal_zone
        self.shot_clock = shot_clock


class Order:
    pass


class Move(Order):
    def __init__(self, velocity: Velocity):
        self.velocity = velocity


class Catch(Order):
    pass


class Kick(Order):
    def __init__(self, velocity: Velocity):
        self.velocity = velocity


class Jump(Order):
    def __init__(self, velocity: Velocity):
        self.velocity = velocity


class OrderSet:
    def __init__(self, turn: int, orders: List[Order], debug_message: str):
        self.turn = turn
        self.orders = orders
        self.debug_message = debug_message


class OrderResponse:
    def __init__(self, code: StatusCode, details: str):
        self.code = code
        self.details = details



from ...protos import physics_pb2
from ...protos import server_pb2

from ...src import specs

from math import floor

# ErrMinCols defines an error for invalid number of cols
ERR_MIN_COLS = AttributeError("number of cols lower the minimum")
# ErrMaxCols defines an error for invalid number of cols
ERR_MAX_COLS = AttributeError("number of cols higher the maximum")
# ErrMinRows defines an error for invalid number of rows
ERR_MIN_ROWS = AttributeError("number of rows lower the minimum")
# ErrMaxRows defines an error for invalid number of rows
ERR_MAX_ROWS = AttributeError("number of rows higher the maximum")

# MinCols Define the min number of cols allowed on the field division by the Mapper
MIN_COLS = 4
# MinRows Define the min number of rows allowed on the field division by the Mapper
MIN_ROWS = 2
# MaxCols Define the max number of cols allowed on the field division by the Mapper
MAX_COLS = 200
# MaxRows Define the max number of rows allowed on the field division by the Mapper
MAX_ROWS = 100


def mirror_coords_to_away(center):
    mirrored = physics_pb2.Point()
    mirrored.x = (specs.MAX_X_COORDINATE - center.x)
    mirrored.y = (specs.MAX_Y_COORDINATE - center.y)
    return mirrored


class Region:
    def __init__(self, col: int, row: int, side: server_pb2.Team.Side, center: physics_pb2.Point, mapper):
        self.col = col
        self.row = row
        self.side = side
        self.center = center
        self.positioner = mapper

    def eq(self, region):
        return region.get_col() == self.col and region.side == self.side and region.get_row() == self.row

    def get_col(self):
        return self.col

    def get_row(self):
        return self.row

    def get_center(self):
        return self.center

    def to_string(self):
        return f"{{{self.col}, {self.row}}}"

    def front(self):
        return self.positioner.get_region(max(self.col + 1, 0), self.row)

    def back(self):
        return self.positioner.get_region(max(self.col - 1, 0), self.row)

    def left(self):
        return self.positioner.get_region(self.col, max(self.row + 1, 0))

    def right(self):
        return self.positioner.get_region(self.col, max(self.row - 1, 0))


class Mapper:
    def __init__(self, cols: int, rows: int, side: server_pb2.Team.Side):
        if cols < MIN_COLS:
            raise ERR_MIN_COLS

        if cols > MAX_COLS:
            raise ERR_MAX_COLS

        if rows < MIN_ROWS:
            raise ERR_MIN_ROWS

        if rows > MAX_ROWS:
            raise ERR_MAX_ROWS

        self.cols = cols
        self.rows = rows
        self.side = side
        self.regionWidth = specs.MAX_X_COORDINATE / cols
        self.regionHeight = specs.MAX_Y_COORDINATE / rows

    def get_region(self, col: int, row: int) -> Region:
        col = max(0, col)
        col = min(self.cols - 1, col)

        row = max(0, row)
        row = min(self.rows - 1, row)

        center = physics_pb2.Point()
        center.x = (round((col * self.regionWidth) + (self.regionWidth / 2)))
        center.y = (round((row * self.regionHeight) + (self.regionHeight / 2)))

        if self.side == server_pb2.Team.Side.AWAY:
            center = mirror_coords_to_away(center)

        return Region(
            col,
            row,
            self.side,
            center,
            self,
        )

    def get_region_from_point(self, point: physics_pb2.Point):
        if self.side == server_pb2.Team.Side.AWAY:
            point = mirror_coords_to_away(point)

        cx = floor(point.x / self.regionWidth)
        cy = floor(point.y / self.regionHeight)
        col = min(cx, self.cols - 1)
        row = min(cy, self.rows - 1)
        return self.get_region(col, row)





# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import server_pb2 as server__pb2


class GameStub(object):
    """Service provided by the game service to the players (clients).
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.JoinATeam = channel.unary_stream(
                '/lugo.Game/JoinATeam',
                request_serializer=server__pb2.JoinRequest.SerializeToString,
                response_deserializer=server__pb2.GameSnapshot.FromString,
                )
        self.SendOrders = channel.unary_unary(
                '/lugo.Game/SendOrders',
                request_serializer=server__pb2.OrderSet.SerializeToString,
                response_deserializer=server__pb2.OrderResponse.FromString,
                )


class GameServicer(object):
    """Service provided by the game service to the players (clients).
    """

    def JoinATeam(self, request, context):
        """JoinATeam allows the player to listen the server during the match.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SendOrders(self, request, context):
        """SendOrders allows the player to send others to the server when the game is on listening state.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_GameServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'JoinATeam': grpc.unary_stream_rpc_method_handler(
                    servicer.JoinATeam,
                    request_deserializer=server__pb2.JoinRequest.FromString,
                    response_serializer=server__pb2.GameSnapshot.SerializeToString,
            ),
            'SendOrders': grpc.unary_unary_rpc_method_handler(
                    servicer.SendOrders,
                    request_deserializer=server__pb2.OrderSet.FromString,
                    response_serializer=server__pb2.OrderResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'lugo.Game', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class Game(object):
    """Service provided by the game service to the players (clients).
    """

    @staticmethod
    def JoinATeam(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/lugo.Game/JoinATeam',
            server__pb2.JoinRequest.SerializeToString,
            server__pb2.GameSnapshot.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SendOrders(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lugo.Game/SendOrders',
            server__pb2.OrderSet.SerializeToString,
            server__pb2.OrderResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



from ..src import lugo
from abc import ABC, abstractmethod


class PlayerState(object):
    SUPPORTING = 0
    HOLDING_THE_BALL = 1
    DEFENDING = 2
    DISPUTING_THE_BALL = 3


PLAYER_STATE = PlayerState()


class Bot(ABC):
    @abstractmethod
    def on_disputing(self, order_set: lugo.OrderSet, snapshot: lugo.GameSnapshot) -> lugo.OrderSet:
        pass

    @abstractmethod
    def on_defending(self, order_set: lugo.OrderSet, snapshot: lugo.GameSnapshot) -> lugo.OrderSet:
        pass

    @abstractmethod
    def on_holding(self, order_set: lugo.OrderSet, snapshot: lugo.GameSnapshot) -> lugo.OrderSet:
        pass

    @abstractmethod
    def on_supporting(self, order_set: lugo.OrderSet, snapshot: lugo.GameSnapshot) -> lugo.OrderSet:
        pass

    @abstractmethod
    def as_goalkeeper(self, order_set: lugo.OrderSet, snapshot: lugo.GameSnapshot, state: PLAYER_STATE) -> lugo.OrderSet:
        pass

    @abstractmethod
    def getting_ready(self, snapshot: lugo.GameSnapshot):
        pass



import os

from . import specs

from ..protos import server_pb2


class EnvVarLoader:

    def __init__(self):
        self._grpcUrl = ""
        self._grpcInsecure = True
        self._botTeamSide = None
        self._botNumber = None
        self._botToken = ""

        if "BOT_TEAM" not in os.environ:
            raise SystemError("missing BOT_TEAM env value")

        if "BOT_NUMBER" not in os.environ:
            raise SystemError("missing BOT_NUMBER env value")

        # the Lugo address
        self._grpcUrl = os.environ.get('BOT_GRPC_URL', 'localhost:5000')
        self._grpcInsecure = bool(os.environ.get('BOT_GRPC_INSECURE', 'false'))

        # defining bot side
        self._botTeamSide = server_pb2.Team.Side.HOME if os.environ[
                                                             "BOT_TEAM"].upper() == 'HOME' else server_pb2.Team.Side.AWAY
        self._botNumber = int(os.environ["BOT_NUMBER"])
        if self._botNumber < 1 or self._botNumber > specs.MAX_PLAYERS:
            raise SystemError('invalid bot number {self._botNumber}, must be between 1 and {specs.MAX_PLAYERS}')

        # // the token is mandatory in official matches, but you may ignore in local games
        self._botToken = os.environ["BOT_TOKEN"] if "BOT_TOKEN" in os.environ else ''

    def get_grpc_url(self):
        return self._grpcUrl

    def get_grpc_insecure(self):
        return self._grpcInsecure

    def get_bot_team_side(self):
        return self._botTeamSide

    def get_bot_number(self):
        return self._botNumber

    def get_bot_token(self):
        return self._botToken



from abc import ABC, abstractmethod
from typing import Callable, Any

from ...src import lugo


#
# The TrainingController is passed to your Training function to give you control of the game flow.
#


class TrainingController(ABC):
    #
    # This method should be called whenever your need to reset the game to an initial state.
    #
    @abstractmethod
    def set_environment(self, data):
        pass

    #
    # Use this method to get the inputs that will be used by your model. E.g. if you are using tensor flow, you may
    # return the tensors used to feed your network.
    #
    @abstractmethod
    def get_state(self):
        pass

    #
    # Use this method to pass that action picked by you model. It will return the reward and `done` values got from
    # your BotTrainer.
    # @param action
    # @returns {Promise<{reward: number, done: boolean}>}
    #
    @abstractmethod
    def update(self, action):
        pass

    #
    # Stops the training
    #

    @abstractmethod
    def stop(self):
        pass


#
# The BotTrainer is used by the Gym class to play the game as a bot and to control the game state when needed.
# It is NOT your learning agent!
#


class BotTrainer(ABC):

    #
    # createNewInitialState should create the initial scenario for each game.
    #
    # IMPORTANT!! Note that this method should define the new state directly on the game server. So you MUST
    # use the remote control client to change the game elements' position/state
    #
    @abstractmethod
    def set_environment(self, data):
        pass

    #
    # getInputs is called in each training step.
    # The training function will call this method to receive whatever inputs you want to use in your neural network.
    # Example, if you have 3 sensors, or 3 tensors, etc.
    #
    # This method must read the game snapshot and define input values and return it
    #
    # @param {GameSnapshot} snapshot - The current game state
    #
    @abstractmethod
    def get_state(self, snapshot: lugo.GameSnapshot):
        pass

    #
    # play define the orders that will be sent to the game server based on the `action` sent by your training function.
    #
    # IMPORTANT:
    # Do not confuse this method role with an agent! Your agent will define the `action` inside your training function.
    # This method is only responsible to translate the action to orders and send them
    #
    # @param {OrderSet} orderSet - used to define the orders that will be sent to the server. Your bot should set the orders
    #                     and return it to the server.
    # @param {GameSnapshot} snapshot - The current game state
    # @param {any} action - Value passed by your training function to the TrainingController `update` method
    #
    #
    @abstractmethod
    def play(self, order_set: lugo.OrderSet, snapshot: lugo.GameSnapshot, action) -> lugo.OrderSet:
        pass

    #
    # This method is called by the TrainingController right after your bot play a turn of the game.
    # It must compare the two states and return the reward and a boolean `done` to indicate that the game each the end.
    #
    # Your bot may evaluate turn by turn, or comparing the final game state to the initial state.
    # However, if you want to compare with the initial state, your bot trainer will have to store the initial
    # state when the method `createNewInitialState` is called.
    #
    # @param {GameSnapshot} previousSnapshot - The current game state
    # @param {GameSnapshot} newSnapshot - The current game state
    #
    @abstractmethod
    def evaluate(self, previous_snapshot: lugo.GameSnapshot, new_snapshot: lugo.GameSnapshot) -> Any:
        pass


TrainingFunction = Callable[[TrainingController], None]



from .src.client import PROTOCOL_VERSION
from .src.client import LugoClient
from .src.client import NewClientFromConfig

from .src.geo import new_vector, sub_vector, is_invalid_vector, get_scaled_vector, get_length, distance_between_points, \
    normalize
from .src.goal import Goal

from .src.interface import Bot, PLAYER_STATE, PlayerState

from .src.loader import EnvVarLoader

from .src.lugo import Order, OrderSet, Point, Vector, new_velocity, Velocity, Team, TeamSide, \
    OrderResponse, \
    CommandResponse, State, Player, PlayerProperties, BallProperties, GameProperties, GameSnapshot, Ball, \
    ResumeListeningResponse, \
    ResumeListeningRequest, PauseResumeRequest, JoinRequest, NextOrderRequest, NextTurnRequest, StatusCode, Jump, Kick, \
    Move, Catch, \
    ShotClock, RemoteServicer

from .src.snapshot import GameSnapshotReader

from .src.specs import *



import asyncio
from .remote_control import RemoteControl
from .interfaces import TrainingController, BotTrainer, TrainingFunction
from ...protos.server_pb2 import GameSnapshot, OrderSet
import threading
from concurrent.futures import ThreadPoolExecutor
import time




class TrainingCrl(TrainingController):

    def __init__(self, executor: ThreadPoolExecutor, remoteControl: RemoteControl, bot: BotTrainer, onReadyCallback: TrainingFunction):
        self._gotNextState = lambda snapshot: print("got first snapshot")
        self.logger = lambda msg: print(f'set debugger')
        self.previousState = None
        self.remoteControl = remoteControl  # type: RemoteControl
        self.onReady = onReadyCallback
        self.trainingHasStarted = False
        self.lastSnapshot = None  # type: GameSnapshot
        self.onListeningMode = False
        self.OrderSet = None
        self.cycleSeq = 0
        self.bot = bot  # type: BotTrainer
        self.debugging_log = False
        self.stopRequested = threading.Event()
        self.trainingExecutor = executor
        self.resumeListeningPhase = lambda action: print(
            'resumeListeningPhase not defined yet - should wait the initialise it on the first "update" call')

    def set_environment(self, data):
        self.logger('Reset state')
        try:
            self.lastSnapshot = self.bot.set_environment(data)
            return self.bot.get_state(self.lastSnapshot)
        except Exception as e:
            print('bot trainer failed to create initial state: ', e)
            raise e

    def get_state(self):
        try:
            self.cycleSeq = self.cycleSeq + 1
            self.logger('get state')
            return self.bot.get_state(self.lastSnapshot)
        except Exception as e:
            print('bot trainer failed to return inputs from a particular state', e)
            raise e

    def update(self, action: any):
        self.logger('received action from training bot')
        if not self.onListeningMode:
            raise ValueError('faulty synchrony - got a new action when was still processing the last one')

        try:
            previousState = self.lastSnapshot
            self.OrderSet.turn = self.lastSnapshot.turn
            updatedOrderSet = self.bot.play(self.OrderSet, self.lastSnapshot, action)

            self.logger('got order set, passing down')

            self.resumeListeningPhase(updatedOrderSet)
            #time.sleep(2.4)  # before calling next turn, let's wait just a bit to ensure the server got our order
            self.lastSnapshot = self.wait_until_next_listening_state()

            self.logger('got new snapshot after order has been sent')

            if self.stopRequested.is_set():
                return None

            # TODO: if I want to skip the net N turns? I should be able too
            self.logger(f"update finished (turn {self.lastSnapshot.turn} waiting for next action)")
        except Exception as e:
            print('failed send new action to the server: ', e)
            raise e

        try:
            return self.bot.evaluate(previousState, self.lastSnapshot)
        except Exception as e:
            print('bot trainer failed to evaluate game state', e)
            raise e

    def gameTurnHandler(self, order_set, snapshot):
        if self.stopRequested.is_set():
            self.logger('skipping turn handler because the stop request')
            return None
        self.logger('new turn')
        if self.onListeningMode:
            raise RuntimeError(
                "faulty synchrony - got new turn while waiting for order (check the lugo 'timer-mode')")

        self._gotNextState(snapshot)
        self.OrderSet = order_set

        waiter = threading.Event()
        new_order_set = None

        def resume(updated_order_set):
            nonlocal new_order_set
            new_order_set = updated_order_set
            waiter.set()
            self.logger(f'Sending new action')

        self.resumeListeningPhase = resume
        self.onListeningMode = True
        if self.trainingHasStarted is False:
            self.trainingExecutor.submit(self.onReady, self, self.stopRequested)
            self.trainingHasStarted = True
            self.logger(f'the training has started')

        self.logger(f'Waiting get new update!')
        waiter.wait(timeout=5)
        self.logger(f'order sent to the game server')
        return new_order_set

    def wait_until_next_listening_state(self) -> GameSnapshot:
        try:
            self.onListeningMode = False
            waiter = threading.Event()

            new_snapshot = None
            def resume(newGameSnapshot):
                nonlocal new_snapshot
                new_snapshot = newGameSnapshot
                waiter.set()

            self._gotNextState = resume


            waiterResumeListening = threading.Event()
            self.trainingExecutor.submit(self.remoteControl.resume_listening, waiterResumeListening)
            waiterResumeListening.wait()

            waiter.wait(timeout=3)
            if new_snapshot is None:
                raise RuntimeError(
                    "timed out waiting for the next listening state - check the training controller")

            self.logger(f'resume_listening: {new_snapshot.turn}')

            return new_snapshot
        except Exception as e:
            self.logger(f'failed to send the orders to the server {e}')
            raise

    def stop(self):
        self.stopRequested.set()





# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: broadcast.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import server_pb2 as server__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0f\x62roadcast.proto\x12\x04lugo\x1a\x0cserver.proto\"\x1e\n\x0eWatcherRequest\x12\x0c\n\x04uuid\x18\x01 \x01(\t\"$\n\x0cStartRequest\x12\x14\n\x0cwatcher_uuid\x18\x01 \x01(\t\"\xfa\x02\n\tGameEvent\x12)\n\rgame_snapshot\x18\x01 \x01(\x0b\x32\x12.lugo.GameSnapshot\x12*\n\nnew_player\x18\x02 \x01(\x0b\x32\x14.lugo.EventNewPlayerH\x00\x12,\n\x0blost_player\x18\x03 \x01(\x0b\x32\x15.lugo.EventLostPlayerH\x00\x12.\n\x0cstate_change\x18\x04 \x01(\x0b\x32\x16.lugo.EventStateChangeH\x00\x12\x1f\n\x04goal\x18\x05 \x01(\x0b\x32\x0f.lugo.EventGoalH\x00\x12(\n\tgame_over\x18\x06 \x01(\x0b\x32\x13.lugo.EventGameOverH\x00\x12\x30\n\nbreakpoint\x18\x07 \x01(\x0b\x32\x1a.lugo.EventDebugBreakpointH\x00\x12\x32\n\x0e\x64\x65\x62ug_released\x18\x08 \x01(\x0b\x32\x18.lugo.EventDebugReleasedH\x00\x42\x07\n\x05\x65vent\"\xfa\x02\n\tGameSetup\x12\x18\n\x10protocol_version\x18\x01 \x01(\t\x12\x10\n\x08\x64\x65v_mode\x18\x02 \x01(\x08\x12\x30\n\nstart_mode\x18\x03 \x01(\x0e\x32\x1c.lugo.GameSetup.StartingMode\x12\x35\n\x0elistening_mode\x18\x04 \x01(\x0e\x32\x1d.lugo.GameSetup.ListeningMode\x12\x1a\n\x12listening_duration\x18\x05 \x01(\r\x12\x15\n\rgame_duration\x18\x06 \x01(\r\x12%\n\thome_team\x18\x08 \x01(\x0b\x32\x12.lugo.TeamSettings\x12%\n\taway_team\x18\t \x01(\x0b\x32\x12.lugo.TeamSettings\"%\n\x0cStartingMode\x12\x0b\n\x07NO_WAIT\x10\x00\x12\x08\n\x04WAIT\x10\x01\"0\n\rListeningMode\x12\t\n\x05TIMER\x10\x00\x12\x08\n\x04RUSH\x10\x01\x12\n\n\x06REMOTE\x10\x02\"N\n\x0cTeamSettings\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0e\n\x06\x61vatar\x18\x02 \x01(\t\x12 \n\x06\x63olors\x18\x03 \x01(\x0b\x32\x10.lugo.TeamColors\"R\n\nTeamColors\x12 \n\x07primary\x18\x01 \x01(\x0b\x32\x0f.lugo.TeamColor\x12\"\n\tsecondary\x18\x02 \x01(\x0b\x32\x0f.lugo.TeamColor\"5\n\tTeamColor\x12\x0b\n\x03red\x18\x01 \x01(\r\x12\r\n\x05green\x18\x02 \x01(\r\x12\x0c\n\x04\x62lue\x18\x04 \x01(\r\".\n\x0e\x45ventNewPlayer\x12\x1c\n\x06player\x18\x01 \x01(\x0b\x32\x0c.lugo.Player\"/\n\x0f\x45ventLostPlayer\x12\x1c\n\x06player\x18\x01 \x01(\x0b\x32\x0c.lugo.Player\"q\n\x10\x45ventStateChange\x12\x30\n\x0eprevious_state\x18\x01 \x01(\x0e\x32\x18.lugo.GameSnapshot.State\x12+\n\tnew_state\x18\x02 \x01(\x0e\x32\x18.lugo.GameSnapshot.State\"*\n\tEventGoal\x12\x1d\n\x04side\x18\x01 \x01(\x0e\x32\x0f.lugo.Team.Side\"\x0f\n\rEventGameOver\"u\n\x14\x45ventDebugBreakpoint\x12\x39\n\nbreakpoint\x18\x01 \x01(\x0e\x32%.lugo.EventDebugBreakpoint.Breakpoint\"\"\n\nBreakpoint\x12\n\n\x06ORDERS\x10\x00\x12\x08\n\x04TURN\x10\x01\"\x14\n\x12\x45ventDebugReleased2\xa8\x01\n\tBroadcast\x12\x32\n\x07OnEvent\x12\x14.lugo.WatcherRequest\x1a\x0f.lugo.GameEvent0\x01\x12\x35\n\x0cGetGameSetup\x12\x14.lugo.WatcherRequest\x1a\x0f.lugo.GameSetup\x12\x30\n\tStartGame\x12\x12.lugo.StartRequest\x1a\x0f.lugo.GameSetupB#Z!github.com/lugobots/lugo4go/protob\x06proto3')



_WATCHERREQUEST = DESCRIPTOR.message_types_by_name['WatcherRequest']
_STARTREQUEST = DESCRIPTOR.message_types_by_name['StartRequest']
_GAMEEVENT = DESCRIPTOR.message_types_by_name['GameEvent']
_GAMESETUP = DESCRIPTOR.message_types_by_name['GameSetup']
_TEAMSETTINGS = DESCRIPTOR.message_types_by_name['TeamSettings']
_TEAMCOLORS = DESCRIPTOR.message_types_by_name['TeamColors']
_TEAMCOLOR = DESCRIPTOR.message_types_by_name['TeamColor']
_EVENTNEWPLAYER = DESCRIPTOR.message_types_by_name['EventNewPlayer']
_EVENTLOSTPLAYER = DESCRIPTOR.message_types_by_name['EventLostPlayer']
_EVENTSTATECHANGE = DESCRIPTOR.message_types_by_name['EventStateChange']
_EVENTGOAL = DESCRIPTOR.message_types_by_name['EventGoal']
_EVENTGAMEOVER = DESCRIPTOR.message_types_by_name['EventGameOver']
_EVENTDEBUGBREAKPOINT = DESCRIPTOR.message_types_by_name['EventDebugBreakpoint']
_EVENTDEBUGRELEASED = DESCRIPTOR.message_types_by_name['EventDebugReleased']
_GAMESETUP_STARTINGMODE = _GAMESETUP.enum_types_by_name['StartingMode']
_GAMESETUP_LISTENINGMODE = _GAMESETUP.enum_types_by_name['ListeningMode']
_EVENTDEBUGBREAKPOINT_BREAKPOINT = _EVENTDEBUGBREAKPOINT.enum_types_by_name['Breakpoint']
WatcherRequest = _reflection.GeneratedProtocolMessageType('WatcherRequest', (_message.Message,), {
  'DESCRIPTOR' : _WATCHERREQUEST,
  '__module__' : 'broadcast_pb2'
  # @@protoc_insertion_point(class_scope:lugo.WatcherRequest)
  })
_sym_db.RegisterMessage(WatcherRequest)

StartRequest = _reflection.GeneratedProtocolMessageType('StartRequest', (_message.Message,), {
  'DESCRIPTOR' : _STARTREQUEST,
  '__module__' : 'broadcast_pb2'
  # @@protoc_insertion_point(class_scope:lugo.StartRequest)
  })
_sym_db.RegisterMessage(StartRequest)

GameEvent = _reflection.GeneratedProtocolMessageType('GameEvent', (_message.Message,), {
  'DESCRIPTOR' : _GAMEEVENT,
  '__module__' : 'broadcast_pb2'
  # @@protoc_insertion_point(class_scope:lugo.GameEvent)
  })
_sym_db.RegisterMessage(GameEvent)

GameSetup = _reflection.GeneratedProtocolMessageType('GameSetup', (_message.Message,), {
  'DESCRIPTOR' : _GAMESETUP,
  '__module__' : 'broadcast_pb2'
  # @@protoc_insertion_point(class_scope:lugo.GameSetup)
  })
_sym_db.RegisterMessage(GameSetup)

TeamSettings = _reflection.GeneratedProtocolMessageType('TeamSettings', (_message.Message,), {
  'DESCRIPTOR' : _TEAMSETTINGS,
  '__module__' : 'broadcast_pb2'
  # @@protoc_insertion_point(class_scope:lugo.TeamSettings)
  })
_sym_db.RegisterMessage(TeamSettings)

TeamColors = _reflection.GeneratedProtocolMessageType('TeamColors', (_message.Message,), {
  'DESCRIPTOR' : _TEAMCOLORS,
  '__module__' : 'broadcast_pb2'
  # @@protoc_insertion_point(class_scope:lugo.TeamColors)
  })
_sym_db.RegisterMessage(TeamColors)

TeamColor = _reflection.GeneratedProtocolMessageType('TeamColor', (_message.Message,), {
  'DESCRIPTOR' : _TEAMCOLOR,
  '__module__' : 'broadcast_pb2'
  # @@protoc_insertion_point(class_scope:lugo.TeamColor)
  })
_sym_db.RegisterMessage(TeamColor)

EventNewPlayer = _reflection.GeneratedProtocolMessageType('EventNewPlayer', (_message.Message,), {
  'DESCRIPTOR' : _EVENTNEWPLAYER,
  '__module__' : 'broadcast_pb2'
  # @@protoc_insertion_point(class_scope:lugo.EventNewPlayer)
  })
_sym_db.RegisterMessage(EventNewPlayer)

EventLostPlayer = _reflection.GeneratedProtocolMessageType('EventLostPlayer', (_message.Message,), {
  'DESCRIPTOR' : _EVENTLOSTPLAYER,
  '__module__' : 'broadcast_pb2'
  # @@protoc_insertion_point(class_scope:lugo.EventLostPlayer)
  })
_sym_db.RegisterMessage(EventLostPlayer)

EventStateChange = _reflection.GeneratedProtocolMessageType('EventStateChange', (_message.Message,), {
  'DESCRIPTOR' : _EVENTSTATECHANGE,
  '__module__' : 'broadcast_pb2'
  # @@protoc_insertion_point(class_scope:lugo.EventStateChange)
  })
_sym_db.RegisterMessage(EventStateChange)

EventGoal = _reflection.GeneratedProtocolMessageType('EventGoal', (_message.Message,), {
  'DESCRIPTOR' : _EVENTGOAL,
  '__module__' : 'broadcast_pb2'
  # @@protoc_insertion_point(class_scope:lugo.EventGoal)
  })
_sym_db.RegisterMessage(EventGoal)

EventGameOver = _reflection.GeneratedProtocolMessageType('EventGameOver', (_message.Message,), {
  'DESCRIPTOR' : _EVENTGAMEOVER,
  '__module__' : 'broadcast_pb2'
  # @@protoc_insertion_point(class_scope:lugo.EventGameOver)
  })
_sym_db.RegisterMessage(EventGameOver)

EventDebugBreakpoint = _reflection.GeneratedProtocolMessageType('EventDebugBreakpoint', (_message.Message,), {
  'DESCRIPTOR' : _EVENTDEBUGBREAKPOINT,
  '__module__' : 'broadcast_pb2'
  # @@protoc_insertion_point(class_scope:lugo.EventDebugBreakpoint)
  })
_sym_db.RegisterMessage(EventDebugBreakpoint)

EventDebugReleased = _reflection.GeneratedProtocolMessageType('EventDebugReleased', (_message.Message,), {
  'DESCRIPTOR' : _EVENTDEBUGRELEASED,
  '__module__' : 'broadcast_pb2'
  # @@protoc_insertion_point(class_scope:lugo.EventDebugReleased)
  })
_sym_db.RegisterMessage(EventDebugReleased)

_BROADCAST = DESCRIPTOR.services_by_name['Broadcast']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z!github.com/lugobots/lugo4go/proto'
  _WATCHERREQUEST._serialized_start=39
  _WATCHERREQUEST._serialized_end=69
  _STARTREQUEST._serialized_start=71
  _STARTREQUEST._serialized_end=107
  _GAMEEVENT._serialized_start=110
  _GAMEEVENT._serialized_end=488
  _GAMESETUP._serialized_start=491
  _GAMESETUP._serialized_end=869
  _GAMESETUP_STARTINGMODE._serialized_start=782
  _GAMESETUP_STARTINGMODE._serialized_end=819
  _GAMESETUP_LISTENINGMODE._serialized_start=821
  _GAMESETUP_LISTENINGMODE._serialized_end=869
  _TEAMSETTINGS._serialized_start=871
  _TEAMSETTINGS._serialized_end=949
  _TEAMCOLORS._serialized_start=951
  _TEAMCOLORS._serialized_end=1033
  _TEAMCOLOR._serialized_start=1035
  _TEAMCOLOR._serialized_end=1088
  _EVENTNEWPLAYER._serialized_start=1090
  _EVENTNEWPLAYER._serialized_end=1136
  _EVENTLOSTPLAYER._serialized_start=1138
  _EVENTLOSTPLAYER._serialized_end=1185
  _EVENTSTATECHANGE._serialized_start=1187
  _EVENTSTATECHANGE._serialized_end=1300
  _EVENTGOAL._serialized_start=1302
  _EVENTGOAL._serialized_end=1344
  _EVENTGAMEOVER._serialized_start=1346
  _EVENTGAMEOVER._serialized_end=1361
  _EVENTDEBUGBREAKPOINT._serialized_start=1363
  _EVENTDEBUGBREAKPOINT._serialized_end=1480
  _EVENTDEBUGBREAKPOINT_BREAKPOINT._serialized_start=1446
  _EVENTDEBUGBREAKPOINT_BREAKPOINT._serialized_end=1480
  _EVENTDEBUGRELEASED._serialized_start=1482
  _EVENTDEBUGRELEASED._serialized_end=1502
  _BROADCAST._serialized_start=1505
  _BROADCAST._serialized_end=1673
# @@protoc_insertion_point(module_scope)



from concurrent.futures import ThreadPoolExecutor

from ...src import client
from ... import mapper

PLAYER_POSITIONS = {
    1: {'Col': 0, 'Row': 1},
    2: {'Col': 1, 'Row': 1},
    3: {'Col': 2, 'Row': 1},
    4: {'Col': 3, 'Row': 1},
    5: {'Col': 4, 'Row': 1},
    6: {'Col': 5, 'Row': 1},
    7: {'Col': 6, 'Row': 1},
    8: {'Col': 7, 'Row': 1},
    9: {'Col': 8, 'Row': 1},
    10: {'Col': 9, 'Row': 1},
    11: {'Col': 10, 'Row': 1},
}


def chaser_turn_handler(team_side, player_number, order_set, snapshot):
    reader = snapshot.GameSnapshotReader(snapshot, team_side)
    order_set.addOrders(reader.make_order_catch())
    me = reader.get_player(team_side, player_number)
    if not me:
        raise ValueError("did not find myself in the game")

    order_set.addOrders(reader.make_order_move_max_speed(
        me.position, snapshot.ball))
    order_set.setDebugMessage(
        f"{'HOME' if team_side == 0 else 'AWAY'}-{player_number} #{snapshot.turn} - chasing ball")
    return order_set


# @background
def newZombieHelperPlayer(team_side, player_number, game_server_address, executor: ThreadPoolExecutor):
    def zombie_turn_handler(order_set, game_snapshot):
        order_set.debug_message = f"{'HOME' if team_side == 0 else 'AWAY'}-{player_number} #{game_snapshot.turn}"
        return order_set

    #
    return newCustomHelperPlayer(team_side, player_number, game_server_address, zombie_turn_handler, executor)


def newChaserHelperPlayer(team_side, player_number, game_server_address):
    return newCustomHelperPlayer(team_side, player_number, game_server_address, chaser_turn_handler)


def newCustomHelperPlayer(team_side, player_number, game_server_address, turn_handler_function,
                          executor: ThreadPoolExecutor):
    try:
        initial_region = mapper.Mapper(22, 5, team_side).get_region(
            PLAYER_POSITIONS[player_number]['Col'], PLAYER_POSITIONS[player_number]['Row'])

        lugo_client = client.LugoClient(
            game_server_address,
            True,
            "",
            team_side,
            player_number,
            initial_region.get_center(),
        )

        def muted():
            None

        lugo_client.play(executor, turn_handler_function, muted)
        return lugo_client
    except Exception as e:
        lugo_client.stop()
        raise e



BASE_UNIT = 100
max_x_coordinate = 200 * BASE_UNIT
max_y_coordinate = 100 * BASE_UNIT
goal_width = 30 * BASE_UNIT
player_max_speed = 100.0

# PLAYER_SIZE is the size of each player
PLAYER_SIZE = 4 * BASE_UNIT

# PLAYER_RECONNECTION_WAIT_TIME is a penalty time imposed to the player that needs to reconnect during the match.
# this interval ensure players won't drop connection in purpose to be reallocated to their initial position.
PLAYER_RECONNECTION_WAIT_TIME = 20

# max number of players in a team by mach
MAX_PLAYERS = 11

# min number of players in a team by mach, if a team gets to have less to this number, the team loses by W.O.
MIN_PLAYERS = 6

# PLAYER_MAX_SPEED is the max speed that a play may move  by frame
PLAYER_MAX_SPEED = player_max_speed

# MAX_X_COORDINATE help us to remember that we should not use the field width to check the field boundaries!
# The field dimensions counts the total number of coordinates, but the coordinates are zero-indexed.
# e.g. If the field width is 10, so the max coordinate will be 9
MAX_X_COORDINATE = max_x_coordinate

# MAX_Y_COORDINATE help us to remember that we should not use the field width to check the field boundaries!
# The field dimensions counts the total number of coordinates, but the coordinates are zero-indexed.
# e.g. If the field height is 10, so the max coordinate will be 9
MAX_Y_COORDINATE = max_y_coordinate

# FIELD_WIDTH is the width of the field (horizontal view)
# This value must be an odd number because the central coordinate must be neutral, and we want to have the same number
# of coordinates on both sides.
# e.g. If the field width is 10, and the coordinates go from 0 to 9, there is no precise middle.
# Thus, the field width would have to be 11, so the coordinate 5 is at the precise center
FIELD_WIDTH = max_x_coordinate + 1

# FIELD_HEIGHT is the height of the field (horizontal view)
# This value must be an odd number because we cant to a coordinate at the perfect middle of the field
# e.g. If the field height is 10, and the coordinates go from 0 to 9, there is no precise middle.
# Thus, the field height would have to be 11, so the coordinate 5 is at the precise center
FIELD_HEIGHT = max_y_coordinate + 1

# FIELD_NEUTRAL_CENTER is the radius of the neutral circle in the center of the field
FIELD_NEUTRAL_CENTER = 1000

# BALL_SIZE size of the element ball
BALL_SIZE = 2 * BASE_UNIT

# BALL_DECELERATION is the deceleration rate of the ball speed  by frame
BALL_DECELERATION = 10.0

# BALL_MAX_SPEED is the max speed of the ball by frame
BALL_MAX_SPEED = 4.0 * BASE_UNIT

# BALL_MIN_SPEED is the minimal speed of the ball  by frame. When the ball was at this speed or slower, it will be considered stopped.
BALL_MIN_SPEED = 2

# BALL_TIME_IN_GOAL_ZONE is the max number of turns that the ball may be in a goal zone. After that, the ball will be auto kicked
# towards the center of the field.
BALL_TIME_IN_GOAL_ZONE = 40 # 40 / 20 fps = 2 seconds,

# GOAL_WIDTH is the goal width
GOAL_WIDTH = goal_width

# GOAL_MIN_Y is the coordinate Y of the lower pole of the goals
GOAL_MIN_Y = (max_y_coordinate - goal_width) / 2

# GOAL_MAX_Y is the coordinate Y of the upper pole of the goals
GOAL_MAX_Y = ((max_y_coordinate - goal_width) / 2) + goal_width

# GOAL_ZONE_RANGE is the minimal distance that a player can stay from the opponent goal
GOAL_ZONE_RANGE = 14 * BASE_UNIT

# GOAL_KEEPER_JUMP_DURATION is the number of turns that the jump takes. A jump cannot be interrupted after has been requested
GOAL_KEEPER_JUMP_DURATION = 3

# GOAL_KEEPER_JUMP_SPEED is the max speed of the goalkeeper during the jump
GOAL_KEEPER_JUMP_SPEED = 2 * player_max_speed

# GOALKEEPER_NUMBER defines the goalkeeper number
GOALKEEPER_NUMBER = 1

# Number of turns each teams has on attack before losing the ball possession.
SHOT_CLOCK_TIME = 300



import grpc
import time
from concurrent.futures import ThreadPoolExecutor
import traceback
from typing import Callable, Iterator

from . import lugo

from ..protos import server_pb2
from ..protos import server_pb2_grpc as server_grpc

from .interface import Bot, PLAYER_STATE
from .loader import EnvVarLoader
from .snapshot import define_state
import threading

PROTOCOL_VERSION = "1.0.0"

RawTurnProcessor = Callable[[lugo.OrderSet, lugo.GameSnapshot], lugo.OrderSet]


# reference https://chromium.googlesource.com/external/github.com/grpc/grpc/+/master/examples/python/async_streaming/client.py
class LugoClient(server_grpc.GameServicer):

    def __init__(self, server_add, grpc_insecure, token, teamSide, number, init_position):
        self._client = None
        self.getting_ready_handler = lambda snapshot: None
        self.callback = Callable[[lugo.GameSnapshot], lugo.OrderSet]
        self.serverAdd = server_add + "?t=" + str(teamSide) + "-" + str(number)
        self.grpc_insecure = grpc_insecure
        self.token = token
        self.teamSide = teamSide
        self.number = number
        self.init_position = init_position
        self._play_finished = threading.Event()
        self._play_routine = None

    def set_client(self, client: server_grpc.GameStub):
        self._client = client

    def get_name(self):
        return f"{'HOME' if self.teamSide == 0 else 'AWAY'}-{self.number}"

    def set_initial_position(self, initial_position: lugo.Point):
        self.init_position = initial_position

    def getting_ready_handler(self, snapshot: lugo.GameSnapshot):
        print(f'Default getting ready handler called for ')

    def set_ready_handler(self, new_ready_handler):
        self.getting_ready_handler = new_ready_handler

    def play(self, executor: ThreadPoolExecutor, callback: Callable[[lugo.GameSnapshot], lugo.OrderSet],
             on_join: Callable[[], None]) -> threading.Event:
        self.callback = callback
        log_with_time(f"{self.get_name()} Starting to play")
        return self._bot_start(executor, callback, on_join)

    def play_as_bot(self, executor: ThreadPoolExecutor, bot: Bot, on_join: Callable[[], None]) -> threading.Event:
        self.set_ready_handler(bot.getting_ready)
        log_with_time(f"{self.get_name()} Playing as bot")

        def processor(orders: lugo.OrderSet, snapshot: lugo.GameSnapshot) -> lugo.OrderSet:
            player_state = define_state(
                snapshot, self.number, self.teamSide)
            if self.number == 1:
                orders = bot.as_goalkeeper(
                    orders, snapshot, player_state)
            else:
                if player_state == PLAYER_STATE.DISPUTING_THE_BALL:
                    orders = bot.on_disputing(orders, snapshot)
                elif player_state == PLAYER_STATE.DEFENDING:
                    orders = bot.on_defending(orders, snapshot)
                elif player_state == PLAYER_STATE.SUPPORTING:
                    orders = bot.on_supporting(orders, snapshot)
                elif player_state == PLAYER_STATE.HOLDING_THE_BALL:
                    orders = bot.on_holding(orders, snapshot)
            return orders

        return self._bot_start(executor, processor, on_join)

    def _bot_start(self, executor: ThreadPoolExecutor, processor: RawTurnProcessor,
                   on_join: Callable[[], None]) -> threading.Event:
        log_with_time(f"{self.get_name()} Starting bot {self.teamSide}-{self.number}")
        if self.grpc_insecure:
            channel = grpc.insecure_channel(self.serverAdd)
        else:
            channel = grpc.secure_channel(
                self.serverAdd, grpc.ssl_channel_credentials())
        try:
            grpc.channel_ready_future(channel).result(timeout=5)
        except grpc.FutureTimeoutError:
            raise Exception(f"timed out waiting to connect to the game server ({self.serverAdd})")

        self.channel = channel
        self._client = server_grpc.GameStub(channel)

        join_request = server_pb2.JoinRequest(
            token=self.token,
            team_side=self.teamSide,
            number=self.number,
            init_position=self.init_position,
        )

        response_iterator = self._client.JoinATeam(join_request)
        on_join()
        self._play_routine = executor.submit(self._response_watcher, response_iterator, processor)
        return self._play_finished

    def stop(self):
        log_with_time(
            f"{self.get_name()} stopping bot - you may need to kill the process if there is no messages coming from "
            f"the server")
        self._play_routine.cancel()
        self._play_finished.set()
        exit(0)

    def wait(self):
        self._play_finished.wait(timeout=None)

    def _response_watcher(
            self,
            response_iterator: Iterator[lugo.GameSnapshot],
            # snapshot,
            processor: RawTurnProcessor) -> None:
        try:
            for snapshot in response_iterator:
                if snapshot.state == lugo.State.OVER:
                    log_with_time(
                        f"{self.get_name()} All done! {lugo.State.OVER}")
                    break
                elif self._play_finished.is_set():
                    break
                elif snapshot.state == lugo.State.LISTENING:
                    orders = server_pb2.OrderSet()
                    orders.turn = snapshot.turn
                    try:
                        orders = processor(orders, snapshot)
                    except Exception as e:
                        traceback.print_exc()
                        log_with_time(f"{self.get_name()}bot processor error: {e}")

                    if orders:
                        self._client.SendOrders(orders)
                    else:
                        log_with_time(
                            f"{self.get_name()} [turn #{snapshot.turn}] bot {self.teamSide}-{self.number} did not return orders")
                elif snapshot.state == lugo.State.GET_READY:
                    self.getting_ready_handler(snapshot)

            self._play_finished.set()
        except grpc.RpcError as e:
            if grpc.StatusCode.INVALID_ARGUMENT == e.code():
                log_with_time(f"{self.get_name()} did not connect {e.details()}")
        except Exception as e:
            log_with_time(f"{self.get_name()} internal error processing turn: {e}")
            traceback.print_exc()


def NewClientFromConfig(config: EnvVarLoader, initialPosition: lugo.Point) -> LugoClient:
    return LugoClient(
        config.get_grpc_url(),
        config.get_grpc_insecure(),
        config.get_bot_token(),
        config.get_bot_team_side(),
        config.get_bot_number(),
        initialPosition,
    )


def log_with_time(msg):
    current_time = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
    print(f"{current_time}: {msg}")



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: physics.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\rphysics.proto\x12\x04lugo\"\x1e\n\x06Vector\x12\t\n\x01x\x18\x01 \x01(\x01\x12\t\n\x01y\x18\x02 \x01(\x01\"\x1d\n\x05Point\x12\t\n\x01x\x18\x01 \x01(\x05\x12\t\n\x01y\x18\x02 \x01(\x05\":\n\x08Velocity\x12\x1f\n\tdirection\x18\x01 \x01(\x0b\x32\x0c.lugo.Vector\x12\r\n\x05speed\x18\x02 \x01(\x01\x42#Z!github.com/lugobots/lugo4go/protob\x06proto3')



_VECTOR = DESCRIPTOR.message_types_by_name['Vector']
_POINT = DESCRIPTOR.message_types_by_name['Point']
_VELOCITY = DESCRIPTOR.message_types_by_name['Velocity']
Vector = _reflection.GeneratedProtocolMessageType('Vector', (_message.Message,), {
  'DESCRIPTOR' : _VECTOR,
  '__module__' : 'physics_pb2'
  # @@protoc_insertion_point(class_scope:lugo.Vector)
  })
_sym_db.RegisterMessage(Vector)

Point = _reflection.GeneratedProtocolMessageType('Point', (_message.Message,), {
  'DESCRIPTOR' : _POINT,
  '__module__' : 'physics_pb2'
  # @@protoc_insertion_point(class_scope:lugo.Point)
  })
_sym_db.RegisterMessage(Point)

Velocity = _reflection.GeneratedProtocolMessageType('Velocity', (_message.Message,), {
  'DESCRIPTOR' : _VELOCITY,
  '__module__' : 'physics_pb2'
  # @@protoc_insertion_point(class_scope:lugo.Velocity)
  })
_sym_db.RegisterMessage(Velocity)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z!github.com/lugobots/lugo4go/proto'
  _VECTOR._serialized_start=23
  _VECTOR._serialized_end=53
  _POINT._serialized_start=55
  _POINT._serialized_end=84
  _VELOCITY._serialized_start=86
  _VELOCITY._serialized_end=144
# @@protoc_insertion_point(module_scope)



import threading

import grpc

from ...src import lugo
from ...protos.remote_pb2 import (
    PauseResumeRequest, NextTurnRequest, NextOrderRequest,
    BallProperties, PlayerProperties, GameProperties,
    ResumeListeningRequest
)
from ...protos.remote_pb2_grpc import RemoteStub


class RemoteControl:
    def __init__(self):
        self.client = None

    def connect(self, grpc_address: str) -> None:
        channel = grpc.insecure_channel(grpc_address)
        try:
            grpc.channel_ready_future(channel).result(timeout=15)
        except grpc.FutureTimeoutError:
            raise Exception("timed out waiting to connect to the game server")
        self.client = RemoteStub(channel)

    def pause_resume(self):
        req = PauseResumeRequest()
        try:
            return self.client.PauseOrResume(req)
        except Exception:
            raise Exception("[Remote Control] Failed to pause/resume the game")

    def resume_listening(self, waiter: threading.Event):
        req = ResumeListeningRequest()
        try:
            result = self.client.ResumeListeningPhase(req)
            waiter.set()
            return result
        except Exception:
            raise Exception("[Remote Control] Failed to resume listening phase the game")

    def next_turn(self):
        req = NextTurnRequest()
        try:
            return self.client.NextTurn(req)
        except Exception:
            raise Exception("[Remote Control] Failed to play to next turn")

    def next_order(self):
        req = NextOrderRequest()
        try:
            return self.client.NextOrder(req)
        except Exception:
            raise Exception("[Remote Control] Failed to play to next order")

    def set_ball_rops(self, position: lugo.Point, velocity: lugo.Velocity):
        req = BallProperties(position=position, velocity=velocity)
        response = self.client.SetBallProperties(req)
        return response

    def set_player_props(self, team_side: lugo.TeamSide, player_number: int, new_position: lugo.Point,
                         new_velocity: lugo.Velocity):
        req = PlayerProperties(
            side=team_side, number=player_number,
            position=new_position, velocity=new_velocity
        )
        response = self.client.SetPlayerProperties(req)
        return response

    def set_game_props(self, turn: int):
        req = GameProperties(turn=turn)
        response = self.client.SetGameProperties(req)
        return response



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import broadcast_pb2 as broadcast__pb2


class BroadcastStub(object):
    """Service to be consumed by clients (e.g. frontend, app, etc) to watch the match.
    The game server implements a Broadcast service. This service may help you to control or watch the game during
    training sessions.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.OnEvent = channel.unary_stream(
                '/lugo.Broadcast/OnEvent',
                request_serializer=broadcast__pb2.WatcherRequest.SerializeToString,
                response_deserializer=broadcast__pb2.GameEvent.FromString,
                )
        self.GetGameSetup = channel.unary_unary(
                '/lugo.Broadcast/GetGameSetup',
                request_serializer=broadcast__pb2.WatcherRequest.SerializeToString,
                response_deserializer=broadcast__pb2.GameSetup.FromString,
                )
        self.StartGame = channel.unary_unary(
                '/lugo.Broadcast/StartGame',
                request_serializer=broadcast__pb2.StartRequest.SerializeToString,
                response_deserializer=broadcast__pb2.GameSetup.FromString,
                )


class BroadcastServicer(object):
    """Service to be consumed by clients (e.g. frontend, app, etc) to watch the match.
    The game server implements a Broadcast service. This service may help you to control or watch the game during
    training sessions.
    """

    def OnEvent(self, request, context):
        """Keep an open stream that publish all important events in the match.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetGameSetup(self, request, context):
        """Returns the game setup configuration.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def StartGame(self, request, context):
        """StartGame allows the master watcher to start the match.
        See the Game Server starting mode to understand how it works.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_BroadcastServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'OnEvent': grpc.unary_stream_rpc_method_handler(
                    servicer.OnEvent,
                    request_deserializer=broadcast__pb2.WatcherRequest.FromString,
                    response_serializer=broadcast__pb2.GameEvent.SerializeToString,
            ),
            'GetGameSetup': grpc.unary_unary_rpc_method_handler(
                    servicer.GetGameSetup,
                    request_deserializer=broadcast__pb2.WatcherRequest.FromString,
                    response_serializer=broadcast__pb2.GameSetup.SerializeToString,
            ),
            'StartGame': grpc.unary_unary_rpc_method_handler(
                    servicer.StartGame,
                    request_deserializer=broadcast__pb2.StartRequest.FromString,
                    response_serializer=broadcast__pb2.GameSetup.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'lugo.Broadcast', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class Broadcast(object):
    """Service to be consumed by clients (e.g. frontend, app, etc) to watch the match.
    The game server implements a Broadcast service. This service may help you to control or watch the game during
    training sessions.
    """

    @staticmethod
    def OnEvent(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/lugo.Broadcast/OnEvent',
            broadcast__pb2.WatcherRequest.SerializeToString,
            broadcast__pb2.GameEvent.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetGameSetup(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lugo.Broadcast/GetGameSetup',
            broadcast__pb2.WatcherRequest.SerializeToString,
            broadcast__pb2.GameSetup.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def StartGame(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lugo.Broadcast/StartGame',
            broadcast__pb2.StartRequest.SerializeToString,
            broadcast__pb2.GameSetup.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)
